#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <strings.h>
#include <string.h>
#include <arpa/inet.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <sys/ioctl.h>

extern double getRealTime();

int     wifiPort = 1111;
char    *serialDevice = "/dev/tty.RaceCapturePro-DevB";

int openSerialDevice ()
{
    int fd; /* File descriptor for the port */
    
    printf("Opening serial port %s...\n", serialDevice);

    fd = open(serialDevice, O_RDWR|O_NOCTTY|O_NDELAY);

    if (fd==-1)
        printf ("Error '%s' opening serial port\n", strerror(errno));
    else
    {
        printf ("Connection established\n");
        fcntl(fd, F_SETFL, 0);
    }
    
    return fd;
}

int main(int argc, char *argv[])
{
    setbuf(stdout, NULL);   //  No buffering for stdout
    
    int     profileInput = 0;
    
    if (argc==1)
        ;
    else if (argc==2&&strcmp(argv[1],"-p")==0)
        profileInput = 1;
    else if (argc==3)
    {
        serialDevice = argv [1];
        wifiPort = atoi (argv[2]);
    }
    else
    {
        printf("Usage: BTSPP2WiFiBridge -p\n");
        printf("    or BTSPP2WiFiBridge\n");
        printf("    or BTSPP2WiFiBridge <btdevice> <port>\n");
        printf("\n");
        printf("With option -p starting a pure input profiling (line per second),\n");
        printf("the standard form using %s as serial device and IP port# %d\n", serialDevice, wifiPort);
        
        exit(1);
    }
    
    printf("Starting BluetoothSPP to WiFi bridge...\n");
    printf("Configuration:\n");
    printf("\tBT SPP device: %s\n", serialDevice);
    
    if (!profileInput)
        printf("\tWiFi port: %d\n", wifiPort);
    
    int     serialFileDescriptor = openSerialDevice ();
   
    if (serialFileDescriptor!=-1)
    {
        if (profileInput)
        {
#define BUFFERSIZE  256
            char                buf [BUFFERSIZE+1];
            
            printf("Sending getMeta initially...\n");
            
            const char          *getMetaCommand = "{\"getMeta\":1}\r\n";
            
            int                 rc = write(serialFileDescriptor,getMetaCommand,strlen(getMetaCommand));
            
            if (rc!=strlen(getMetaCommand))
                printf("Error condition when writing getMeta to serial port...\n");
            
            printf("Profiling...\n");

            int                 lineCount = 0;
            double              startTime = getRealTime();
            
            //  Read from serial port
            while (1)
            {
                rc = read(serialFileDescriptor, &buf, BUFFERSIZE);
                
                if (rc==0)
                {
                    printf("Serial port closed connection\n");
                    break;
                }
                else if (rc<0)
                {
                    printf("Error condition in when reading from serial port\n");
                    break;
                }
                else
                {
                    printf ("BT -> IP: forwarded '");
                    
                    for (int i = 0; i<rc; i++)
                        switch (buf [i])
                        {
                            case '\r':
                                
                                printf ("\\r");
                                break;
                                
                            case '\n':
                                
                                printf ("\\n");
                                lineCount++;
                                break;
                                
                            default:
                                
                                printf ("%c", buf [i]);
                                break;
                        }
                    printf ("'\n");
                }
                
                if (lineCount>1000)
                {
                    double      updateRate = lineCount/(getRealTime ()-startTime);
                    
                    printf("Update rate measured %.1f Hz\n", updateRate);
                    break;
                }
            }
        }
        else
        {
            /* Socket erstellen - TCP, IPv4, keine Optionen */
            int     lsd = socket(AF_INET, SOCK_STREAM, 0);
         
            /* IPv4, Port: wifiPort, jede IP-Adresse akzeptieren */
            struct sockaddr_in saddr;
            
            saddr.sin_family = AF_INET;
            saddr.sin_port = htons(wifiPort);
            saddr.sin_addr.s_addr = htons(INADDR_ANY);
         
            int err = -1;
            /* Socket an Port binden */
            do
            {
                err = bind (lsd, (struct sockaddr*) &saddr, sizeof(saddr));
            
                if (err==0)
                {
                    /* Auf Socket horchen (Listen) */
                    err = listen(lsd, 10);
                    
                    if (err==-1)
                        printf("Error listening to port# %d (%d), sleeping...\n", wifiPort, errno);
                }
                else
                    printf("Error binding to port# %d (%d), sleeping...\n", wifiPort, errno);
                
                if (err==-1)
                    sleep (10);
            }
            while (err!=0);
        
            int     runmainloop = 1;
            
            while(runmainloop)
            {
                /* Puffer und Strukturen anlegen */
                struct sockaddr_in  clientaddr;
                char                buffer[10];
                int                 numStatus = 0;
                
                bzero(buffer, sizeof(buffer));
         
                printf("Listening on network port# 1111...\n");

                /* Auf Verbindung warten, bei Verbindung Connected-Socket erstellen */
                socklen_t           clen = sizeof(clientaddr);
                int                 csd = accept(lsd, (struct sockaddr*) &clientaddr, &clen);

                if (csd>=0)
                {
                    printf("Network connection established\n");
                    
                    fcntl(csd, F_SETFL, fcntl(csd, F_GETFL, 0) | O_NONBLOCK);
                }
                else
                    printf("Network connecting failed\n");

#define BUFFERSIZE  256
                char                buf [BUFFERSIZE+1];

                while (csd>=0)
                {
                    int     bytesAvailable = 0;
                    int     rc = ioctl(serialFileDescriptor, FIONREAD, &bytesAvailable);
                    
                    //  Check BT SPP to IP port...
                    if (rc!=-1&&bytesAvailable>0)
                    {
                        //  Read from serial port and send input to network connection
                        rc = read(serialFileDescriptor, buf, BUFFERSIZE);

                        if (rc==0)
                        {
                            printf("Serial port closed connection\n");
                            runmainloop = 0;
                            break;
                        }
                        else if (rc<0)
                        {
                            printf("Error condition '%s' in when reading from serial port\n", strerror(errno));
                            runmainloop = 0;
                            break;
                        }
                        else
                        {
                            int rc2 = send (csd, buf, rc, 0);
                        
                            if (rc2==0)
                            {
                                printf("BT -> IP: network connection closed...\n");
                                break;
                            }
                            else if (rc2<0)
                            {
                                printf("BT -> IP: error condition writing to network connection\n");
                                break;
                            }
                            else if (rc!=rc2)
                                printf("BT -> IP: incomplete write to IP port!\n");
                            else
                            {
                                printf ("BT -> IP: forwarded '");
                                
                                for (int i = 0; i<rc2; i++)
                                    switch (buf [i])
                                    {
                                        case '\r':
                                            
                                            printf ("\\r");
                                            break;
                                            
                                        case '\n':
                                            
                                            printf ("\\n");
                                            break;
                                            
                                        default:
                                            
                                            printf ("%c", buf [i]);
                                            break;
                                    }
                                
                                printf ("'\n");
                            }
                        }
                    }
                    
                    //  Check IP port to BT SPP...
                    bytesAvailable = 0;
                    
                    rc = ioctl(csd, FIONREAD, &bytesAvailable);
                    
                    //  Check BT SPP to IP port...
                    if (rc!=-1&&bytesAvailable>0)
                    {
                        rc = recv (csd, buf, BUFFERSIZE, 0);
                    
                        if (rc>0)
                        {
                            int     rc2 = write(serialFileDescriptor,buf,rc);
                            
                            if (rc2<0)
                            {
                                printf("IP -> BT: error condition '%s' when writing to serial port\n", strerror(errno));
                                runmainloop = 0;
                                break;
                            }
                            else if (rc2!=rc)
                                printf("IP -> BT: incomplete write to serial port!\n");
                            else
                            {
                                printf ("IP -> BT: forwarded '");
                                
                                for (int i = 0; i<rc2; i++)
                                    switch (buf [i])
                                {
                                    case '\r':
                                        
                                        printf ("\\r");
                                        break;
                                        
                                    case '\n':
                                        
                                        printf ("\\n");
                                        break;
                                        
                                    default:
                                        
                                        printf ("%c", buf [i]);
                                        break;
                                }
                                
                                printf ("'\n");
                            }
                            
                        }
                        else if (rc==0)
                        {
                            printf("IP port closed connection\n");
                            runmainloop = 0;
                            break;
                        }
                        else
                        {
                            printf("Error condition '%s' when reading from IP port\n", strerror(errno));
                            runmainloop = 0;
                            break;
                        }
                    }
                }
                
                close(csd);
            }
        }
        
        close (serialFileDescriptor);
    }
    
	return EXIT_SUCCESS;
}

